class PHPParser extends Parser;

options {
    importVocab = PHPScanner;
    //codeGenMakeSwitchThreshold = 2;
    //codeGenBitsetTestThreshold = 6;
    buildAST = true;
    k = 1;
}

tokens {
    T_PRE_INC;
    T_PRE_DEC;
    T_POST_INC;
    T_POST_DEC;
    T_UNARY_MINUS;
    T_UNARY_PLUS;
    T_FUNCTION_CALL;
    T_BLOCK;
    T_FOR_INIT;
    T_FOR_TEST;
    T_FOR_ACTION;
    T_FOREACH_ARRAY;
    T_FOREACH_KEY;
    T_FOREACH_VALUE;
    T_METHOD_DECLARATION;
    T_FIELD_DECLARATION;
    T_MEMBER_MODIFIER_LIST;
    T_CLASS_VARIABLE_INITIALIZATION;
    T_CLASS_BODY;
}

start:
        inner_statement_list
;

top_statement:
                statement
        |       function_declaration_statement
        |       class_declaration_statement
;

inner_statement_list:
        (inner_statement)*
;

inner_statement:
                statement
        |       function_declaration_statement
        |       class_declaration_statement
;

block_statement:
        o:T_OPEN_CURLY^ inner_statement_list T_CLOSE_CURLY! {#o.setType(T_BLOCK);}
;

new_block_statement:
        c:T_COLON^ inner_statement_list {#c.setType(T_BLOCK);}
;

statement:
                block_statement
        |       T_IF^ T_OPEN_ROUND! expression T_CLOSE_ROUND! (
                                (new_block_statement (new_elseif)* (new_else_single)? T_ENDIF! T_SEMI_COLON!)
                        |       (statement (elseif)* (else_single)?)
                )
        //|       T_IF T_OPEN_ROUND expression T_CLOSE_ROUND T_COLON inner_statement_list new_elseif_list new_else_single T_ENDIF T_SEMI_COLON
        |       T_WHILE^ T_OPEN_ROUND! expression T_CLOSE_ROUND! while_statement
        |       T_DO^ statement T_WHILE! T_OPEN_ROUND! expression T_CLOSE_ROUND! T_SEMI_COLON!
        |       T_FOR^
                        for_init
                        for_test
                        for_action
                        //T_OPEN_ROUND! for_expr T_SEMI_COLON! for_expr T_SEMI_COLON! for_expr
                        T_CLOSE_ROUND!
                        for_statement
        |       T_FOREACH^ 
                        T_OPEN_ROUND! (expression)
                        //(T_AS^ is_reference variable (T_DOUBLE_ARROW^ is_reference variable)?) T_CLOSE_ROUND!
                        //foreach_array
                        //foreach_key
                        //(foreach_value)?
                        foreach_iterator
                        T_CLOSE_ROUND!
                        foreach_statement
        |       T_SWITCH^ T_OPEN_ROUND! expression T_CLOSE_ROUND! switch_case_list
        |       T_BREAK^ (expression)? T_SEMI_COLON!
        |       T_CONTINUE^ (expression)? T_SEMI_COLON!
        |       T_RETURN^ (expression)? T_SEMI_COLON!
        |       T_GLOBAL^ global_var_list T_SEMI_COLON!
        |       T_STATIC^ static_var_list T_SEMI_COLON!
        |       T_ECHO^ echo_expr_list T_SEMI_COLON!
        |       T_INLINE_HTML^
        |       expression T_SEMI_COLON!
        |       T_USE^ use_filename T_SEMI_COLON!
        |       T_UNSET^ T_OPEN_ROUND! unset_variables T_CLOSE_ROUND! T_SEMI_COLON!
        |       T_DECLARE^ T_OPEN_ROUND! declare_list T_CLOSE_ROUND! declare_statement
        |       T_SEMI_COLON!            /* empty statement */
        |       T_TRY^ block_statement
                (T_CATCH^ T_OPEN_ROUND! T_STRING T_VARIABLE T_CLOSE_ROUND! block_statement)+
        |       T_THROW^ expression T_SEMI_COLON!
;

unset_variables:
                variable (T_COMMA! variable)*
;

use_filename:
                T_CONSTANT_ENCAPSED_STRING
        |       T_OPEN_ROUND! T_CONSTANT_ENCAPSED_STRING T_CLOSE_ROUND!
;

function_declaration_statement:
                unticked_function_declaration_statement
;

class_declaration_statement:
                unticked_class_declaration_statement
;

reference_variable:
                T_AMPERSAND^ variable
;

is_reference:
                (T_AMPERSAND^)?
;

unticked_function_declaration_statement:
                T_FUNCTION^ is_reference T_STRING
                function_declaration_parameter_list
                function_body
;

function_declaration_parameter_list:
                T_OPEN_ROUND^ parameter_list T_CLOSE_ROUND!
;

function_body:
                block_statement
;

unticked_class_declaration_statement:
                (T_ABSTRACT | T_FINAL)? T_CLASS^ T_STRING
                        (class_declaration_extends)?
                        (class_declaration_implements)?
                        class_declaration_body
        |       T_INTERFACE^ T_STRING (interface_declaration_extends)? class_declaration_body
;

interface_declaration_extends:
        T_EXTENDS^ interface_list
;

class_declaration_extends:
        T_EXTENDS^ T_STRING
;

class_declaration_implements:
        T_IMPLEMENTS^ interface_list
;

class_declaration_body:
        o:T_OPEN_CURLY^ class_statement_list T_CLOSE_CURLY! {#o.setType(T_CLASS_BODY);}
;
    

interface_list:
        T_STRING (T_COMMA! T_STRING)*
;

for_statement:
                statement
        |       new_block_statement T_ENDFOR! T_SEMI_COLON!
;

foreach_statement:
                statement
        |       new_block_statement T_ENDFOREACH! T_SEMI_COLON!
;

declare_statement:
                statement
        |       new_block_statement T_ENDDECLARE! T_SEMI_COLON!
;


declare_list:
                T_STRING T_EQUAL static_scalar (T_COMMA! T_STRING T_EQUAL static_scalar)*
;

switch_case_list:
                o:T_OPEN_CURLY^ (T_SEMI_COLON!)? case_list T_CLOSE_CURLY! {#o.setType(T_BLOCK); }
        |       c:T_COLON^ (T_SEMI_COLON!)? case_list T_ENDSWITCH! T_SEMI_COLON! {#c.setType(T_BLOCK);}
;

case_list: // XXX differs form original, only one default allowed, must be last
                (case_case)*
                (case_default)?
;

case_case:
        T_CASE^ expression case_action
;

case_action:
                c:T_COLON^ inner_statement_list {#c.setType(T_BLOCK);}
        |       s:T_SEMI_COLON^ inner_statement_list {#s.setType(T_BLOCK);}
        //(T_COLON^ | T_SEMI_COLON^) inner_statement_list
;

case_default:
        //T_DEFAULT^ (T_COLON! | T_SEMI_COLON!) inner_statement_list
        T_DEFAULT^ case_action
;

while_statement:
                statement
        |       new_block_statement T_ENDWHILE! T_SEMI_COLON!
;

elseif:
                T_ELSEIF^ T_OPEN_ROUND! expression T_CLOSE_ROUND! statement
;

new_elseif:
                T_ELSEIF^ T_OPEN_ROUND! expression T_CLOSE_ROUND! new_block_statement
;

else_single:
                T_ELSE^ statement
;


new_else_single:
                T_ELSE^ new_block_statement
;

parameter_list:
                (non_empty_parameter_list)?
;


non_empty_parameter_list:
        (
            (T_STRING)? // optional_class_type
            function_parameter
        )
        (
                T_COMMA!
                (T_STRING)? // optional_class_type
                function_parameter
        )*
;

function_parameter:
            (T_AMPERSAND^)? T_VARIABLE (T_EQUAL^ static_scalar)?
;

function_call_parameter_list:
                (non_empty_function_call_parameter_list)?
;

non_empty_function_call_parameter_list:
                expression (T_COMMA! expression)*
;

global_var_list:
                global_var (T_COMMA! global_var)*
;

global_var:
                T_VARIABLE
        |       T_DOLLAR (variable | T_OPEN_CURLY expression T_CLOSE_CURLY)
;

static_var_list:
                T_VARIABLE (T_EQUAL static_scalar)? (T_COMMA! T_VARIABLE (T_EQUAL static_scalar)?)*
;

class_statement_list:
                (class_statement)*
;

class_statement:
                (member_modifier_list T_FUNCTION) => m1:member_modifier_list! method_declaration[#m1] //T_FUNCTION^ is_reference T_STRING function_declaration_parameter_list method_body
        |       m2:non_empty_member_modifier_list! class_variable_declaration[#m2]
        //|       m3:T_VAR! class_variable_declaration[#m3] // deprecated with E_STRICT
        |       T_VAR! class_variable_declaration[#([T_PUBLIC, "public"])] // deprecated with E_STRICT, same as public
        |       class_constant_declaration
;

method_declaration![AST modifiers]:
                T_FUNCTION ref:is_reference name:T_STRING param:function_declaration_parameter_list body:method_body
                {#method_declaration = #([T_METHOD_DECLARATION], name, modifiers, ref, param, body);}
;

method_body:
                T_SEMI_COLON! /* abstract method */
        |       block_statement
;

member_modifier_list:
                //(member_modifier)* {#member_modifier_list = #([T_MEMBER_MODIFIER_LIST], #member_modifier_list);}
                (non_empty_member_modifier_list)?
;

non_empty_member_modifier_list:
                (member_modifier)+ {#non_empty_member_modifier_list = #([T_MEMBER_MODIFIER_LIST], #non_empty_member_modifier_list);}
;

member_modifier:
                (T_PUBLIC | T_PROTECTED | T_PRIVATE | T_STATIC | T_ABSTRACT | T_FINAL)
;

class_variable_declaration![AST modifiers]:
         init:class_variable_initialization_list! T_SEMI_COLON! {#class_variable_declaration = #([T_CLASS_VARIABLE_INITIALIZATION, "var"], modifiers, init);}
;

class_variable_initialization_list:
        class_variable_initialization (T_COMMA! class_variable_initialization)*
;

class_variable_initialization:
        T_VARIABLE (T_EQUAL! static_scalar)? {#class_variable_initialization = #([T_CLASS_VARIABLE_INITIALIZATION], #class_variable_initialization);}
;

class_constant_declaration:
        T_CONST^ class_constant_initialization_list T_SEMI_COLON!
        //T_CONST^ T_STRING T_EQUAL^ static_scalar (T_COMMA! T_STRING T_EQUAL static_scalar)* T_SEMI_COLON!
;

class_constant_initialization:
        T_STRING T_EQUAL^ static_scalar
;

class_constant_initialization_list:
        class_constant_initialization (T_COMMA! class_constant_initialization)*
;

echo_expr_list:
                expression (T_COMMA! expression)*
;

/*
foreach_array:
        x:T_OPEN_ROUND! expression {#x.setType(T_FOREACH_ARRAY);}
;

foreach_key:
        //(T_AS^ is_reference variable (T_DOUBLE_ARROW^ is_reference variable)?) T_CLOSE_ROUND!
        x:T_AS^ reference_variable {#x.setType(T_FOREACH_KEY);}

;

foreach_value:
        x:T_DOUBLE_ARROW^ reference_variable {#x.setType(T_FOREACH_VALUE);}
;
*/

foreach_iterator:
        T_AS! (reference_variable | variable) (T_DOUBLE_ARROW^ (reference_variable | variable))?
;

for_init:
        x:T_OPEN_ROUND^ for_expr {#x.setType(T_FOR_INIT);}
;

for_test:
        x:T_SEMI_COLON^ for_expr {#x.setType(T_FOR_TEST);}
;

for_action:
        x:T_SEMI_COLON^ for_expr {#x.setType(T_FOR_ACTION);}
;

for_expr:
        (non_empty_for_expr)?
;

non_empty_for_expr:
                expression (T_COMMA! expression)*
;

expression:
                logical_or_expression
        |       internal_functions
;

/*
expression_list:
        expression (T_COMMA! expression)*
;
*/

logical_or_expression:
        logical_xor_expression (T_LOGICAL_OR^ logical_xor_expression)*
;

logical_xor_expression:
        logical_and_expression (T_LOGICAL_XOR^ logical_and_expression)*
;

logical_and_expression:
        print_expression (T_LOGICAL_AND^ print_expression)*
;

print_expression:
                T_PRINT^ print_expression
        |       assignment_expression
;

assignment_expression:
                ((T_NOT)* assignment) => (T_NOT^)* assignment
        |       ternary_expression
;

assignment:
        (variable (T_EQUAL^ | T_PLUS_EQUAL^ | T_MINUS_EQUAL^ | T_MUL_EQUAL^ | T_DIV_EQUAL^ | T_CONCAT_EQUAL^ | T_MOD_EQUAL^ | T_OR_EQUAL^ | T_XOR_EQUAL^ | T_SL_EQUAL^ | T_SR_EQUAL^)) assignment_expression
;

ternary_expression:
        // not sure about precedence
        boolean_or_expression (T_QUESTION^ boolean_or_expression T_COLON! boolean_or_expression)*
;

boolean_or_expression:
        boolean_and_expression (T_BOOLEAN_OR^ boolean_and_expression)*
;

boolean_and_expression:
        binary_or_expression (T_BOOLEAN_AND^ binary_or_expression)*
;

binary_or_expression:
        binary_xor_expression (T_PIPE^ binary_xor_expression)*
;

binary_xor_expression:
        binary_and_expression (T_EXP^ binary_and_expression)*
;

binary_and_expression:
        equal_expression (T_AMPERSAND^ equal_expression)*
;

equal_expression:
        compare_expression ((T_IS_EQUAL^ | T_IS_NOT_EQUAL^ | T_IS_IDENTICAL^ | T_IS_NOT_IDENTICAL^) compare_expression)?
;

compare_expression:
        shift_expression ((T_LT^ | T_IS_SMALLER_OR_EQUAL^ | T_GT^ | T_IS_GREATER_OR_EQUAL^) shift_expression)?
;

shift_expression:
        add_expression ((T_SL^ | T_SR^) add_expression)*
;

add_expression:
        mul_expression ((T_PLUS^ | T_MINUS^ | T_DOT^) mul_expression)*
;

mul_expression:
        logical_not_expression ((T_MUL^ | T_DIV^ | T_MOD^) logical_not_expression)*
;

logical_not_expression:
                T_NOT^ logical_not_expression
        |       instance_of_expression
;

instance_of_expression:
        unary_expression (T_INSTANCEOF^ class_name_reference)?
;

unary_expression:
                (T_TILDE^ | T_INT_CAST^ | T_DOUBLE_CAST^ | T_STRING_CAST^ | T_ARRAY_CAST^ | T_OBJECT_CAST^ | T_UNSET_CAST^ | T_AT^) unary_expression
        |       p:T_PLUS!  unary_expression // {#p.setType(T_UNARY_PLUS);} // unary plus has no effect
        |       m:T_MINUS^ unary_expression {#m.setType(T_UNARY_MINUS);}
        |       i:T_INC^   variable         {#i.setType(T_PRE_INC);}
        |       d:T_DEC^   variable         {#d.setType(T_PRE_DEC);}
        |       new_expression
;

new_expression:
                (T_AMPERSAND T_NEW) => T_AMPERSAND! T_NEW^ class_name_reference (ctor_arguments)? //(T_OPEN_ROUND^ function_call_parameter_list T_CLOSE_ROUND!)?
        |       T_NEW^ class_name_reference (ctor_arguments)? //(T_OPEN_ROUND! function_call_parameter_list T_CLOSE_ROUND!)?
        //|       (T_AMPERSAND!)? T_NEW^ class_name_reference (ctor_arguments)? //(T_OPEN_ROUND! function_call_parameter_list T_CLOSE_ROUND!)?
        |       T_CLONE^ new_expression
        |       scalar_expression
;

scalar_expression:
                common_scalar
        |       variable (
                                i2:T_INC^ {#i2.setType(T_POST_INC);}
                         |      d2:T_DEC^ {#d2.setType(T_POST_DEC);}
                )?
        |       T_AMPERSAND^ variable
        |       T_STRING_VARNAME^
        |       T_DOUBLEQUOTE^ encaps_list T_DOUBLEQUOTE!
        |       T_SINGLEQUOTE^ encaps_list T_SINGLEQUOTE!
        |       T_BACKQUOTE^ encaps_list T_BACKQUOTE!
        |       T_START_HEREDOC^ encaps_list T_END_HEREDOC!
        |       T_ARRAY^ T_OPEN_ROUND! array_pair_list T_CLOSE_ROUND!
        |       T_LIST^ T_OPEN_ROUND! assignment_list T_CLOSE_ROUND!
        |       T_OPEN_ROUND! expression T_CLOSE_ROUND!
;

class_name_reference:
                T_STRING (T_PAAMAYIM_NEKUDOTAYIM variable_without_objects (T_OBJECT_OPERATOR object_property)*)?
        |       variable_without_objects (T_OBJECT_OPERATOR object_property)*
;

ctor_arguments:
                T_OPEN_ROUND^ function_call_parameter_list T_CLOSE_ROUND!
;

common_scalar:
                T_LNUMBER
        |       T_DNUMBER
        |       T_CONSTANT_ENCAPSED_STRING
        |       T_LINE
        |       T_FILE
        |       T_CLASS_C
        |       T_METHOD_C
        |       T_FUNC_C
;

static_scalar: /* compile-time evaluated scalars */
                common_scalar
        |       T_STRING (T_PAAMAYIM_NEKUDOTAYIM^ T_STRING)?
        |       T_PLUS^ static_scalar
        |       T_MINUS^ static_scalar
        |       T_ARRAY^ T_OPEN_ROUND! static_array_pair_list T_CLOSE_ROUND!
;


/*
scalar:
                T_STRING (T_PAAMAYIM_NEKUDOTAYIM T_STRING)? // class_constant
        |       T_STRING_VARNAME
        |       common_scalar
        |       T_DOUBLEQUOTE encaps_list T_DOUBLEQUOTE
        |       T_SINGLEQUOTE encaps_list T_SINGLEQUOTE
        |       T_START_HEREDOC encaps_list T_END_HEREDOC
;
*/

variable:
        (variable_without_objects | T_STRING )
        (
                T_PAAMAYIM_NEKUDOTAYIM^ (
                                variable_without_objects
                        |       T_STRING
                )
        )?
        (T_OPEN_ROUND^ function_call_parameter_list T_CLOSE_ROUND!)?
        (T_OBJECT_OPERATOR^ object_property (T_OPEN_ROUND^ function_call_parameter_list T_CLOSE_ROUND!)?)*
;

/*
object_property_reference:
        (T_OBJECT_OPERATOR^ object_property method_or_not)*
;
*/

method_or_not:
        (T_OPEN_ROUND^ function_call_parameter_list T_CLOSE_ROUND!)?
;

variable_without_objects:
        (
                        T_VARIABLE
                |       (
                            (T_DOLLAR^)+
                            (
                                            T_VARIABLE
                                    |       (T_OPEN_CURLY^ expression T_CLOSE_CURLY!)
                            )
                        )
        )
        (
                        (T_OPEN_SQUARE^ dim_offset T_CLOSE_SQUARE!)
                |       (T_OPEN_CURLY^ expression T_CLOSE_CURLY!)
        )*
;

dim_offset:
                (expression)?
;

object_property:
                object_dim_list
        |       variable_without_objects
;

object_dim_list:
                variable_name
                (
                                T_OPEN_SQUARE^ dim_offset T_CLOSE_SQUARE!
                        |       T_OPEN_CURLY^ expression T_CLOSE_CURLY!
                )*
;

variable_name:
                T_STRING
        |       T_OPEN_CURLY^ expression T_CLOSE_CURLY!
;

simple_indirect_reference:
                (T_DOLLAR)+
;

assignment_list:
                assignment_list_element (T_COMMA! assignment_list_element)*
;

assignment_list_element:
        (
                        variable
                |       (T_LIST^ T_OPEN_ROUND! assignment_list T_CLOSE_ROUND!)
        )?
;

static_array_pair_list:
        (non_empty_static_array_pair_list)?
;

non_empty_static_array_pair_list:
        //static_array_pair (T_COMMA! static_array_pair)* (T_COMMA)?
        //static_array_pair ((T_COMMA T_CLOSE_ROUND) => T_COMMA | (T_COMMA static_array_pair)*)
        //static_array_pair ((T_COMMA static_array_pair) => (T_COMMA static_array_pair)* | (T_COMMA))
        static_array_pair ((T_COMMA static_array_pair) => (T_COMMA! static_array_pair))* (T_COMMA!)?
;

static_array_pair:
        static_scalar (T_DOUBLE_ARROW^ static_scalar)?
;

array_pair_list:
        (non_empty_array_pair_list)?
        //(array_pair (T_COMMA! array_pair)* (T_COMMA!)?)?
;

non_empty_array_pair_list:
        //array_pair (T_COMMA! array_pair)* (T_COMMA!)?
        array_pair ((T_COMMA array_pair) => (T_COMMA! array_pair))* (T_COMMA!)?
        //array_pair (((T_COMMA T_CLOSE_ROUND) => T_COMMA)? | (T_COMMA array_pair)*)
;

array_pair:
        expression (T_DOUBLE_ARROW^ expression)?
;

encaps_list:
        (
                        encaps_var
                |       T_STRING
                |       T_NUM_STRING
                |       T_ENCAPSED_AND_WHITESPACE
                |       T_CHARACTER
                |       T_BAD_CHARACTER
                |       T_OPEN_SQUARE
                |       T_CLOSE_SQUARE
                |       T_OPEN_CURLY
                |       T_CLOSE_CURLY
                |       T_OBJECT_OPERATOR
        )*
;

encaps_var:
                T_VARIABLE (
                                (T_OBJECT_OPERATOR T_STRING) => T_OBJECT_OPERATOR^ T_STRING 
                        |       (T_OPEN_SQUARE encaps_var_offset T_CLOSE_SQUARE) => T_OPEN_SQUARE^ encaps_var_offset T_CLOSE_SQUARE!
                )?
        |       T_DOLLAR_OPEN_CURLY_BRACES^ ((T_STRING_VARNAME T_OPEN_SQUARE expression) => (T_STRING_VARNAME T_OPEN_SQUARE^ expression T_CLOSE_SQUARE!) | expression) T_CLOSE_CURLY!
        |       T_CURLY_OPEN^ variable T_CLOSE_CURLY!
;

encaps_var_offset:
                T_STRING
        |       T_NUM_STRING
        |       T_VARIABLE
;

internal_functions:
                T_ISSET^ T_OPEN_ROUND! isset_variables T_CLOSE_ROUND!
        |       T_EMPTY^ T_OPEN_ROUND! variable T_CLOSE_ROUND!
        |       T_INCLUDE^ expression
        |       T_INCLUDE_ONCE^ expression
        |       T_EVAL^ T_OPEN_ROUND! expression T_CLOSE_ROUND!
        |       T_REQUIRE^ expression
        |       T_REQUIRE_ONCE^ expression
        |       T_EXIT^ (T_OPEN_ROUND! (expression)? T_CLOSE_ROUND!)?
;

isset_variables:
                variable (T_COMMA! variable)*
;

class_constant:
                T_STRING T_PAAMAYIM_NEKUDOTAYIM^ T_STRING
;
