/**
 * PHP Scanner
 */

%%

%{
    private static final boolean ENABLE_ASP_TAGS     = false;
    private static final boolean ENABLE_SHORT_TAGS   = true;

    private java.util.Stack  zzLexicalStateStack = new java.util.Stack();
    private StringBuffer     comment             = new StringBuffer();
    private String           heredocLabel;

    public final void yypushstate(int newState)
    {
      zzLexicalStateStack.push(new Integer(zzLexicalState));
      zzLexicalState = newState;
    }
   
    public final void yypopstate()
    {
      zzLexicalState = ((Integer)zzLexicalStateStack.pop()).intValue();
    }

    private Token newToken(int index)
    {
        //return new Token(index, yyline, yycolumn);
        return new Token(index, yyline, yycolumn, yytext());
    }

    private Token newTokenText(int index, String text)
    {
        return new Token(index, yyline, yycolumn, yytext(), text);
    }

    private Token newTokenText(int index)
    {
        return new Token(index, yyline, yycolumn, yytext(), yytext());
    }

    private Token newTokenInt(int index, String text)
    {
        int i;
        try {
            i = Integer.parseInt(text, 10);
        }
        catch (NumberFormatException nfe) {
            return newTokenDouble(index, text);
        }
        return new Token(index, yyline, yycolumn, yytext(), new Integer(i));
    }

    private Token newTokenInt(int index)
    {
        return newTokenInt(index, yytext());
    }

    private Token newTokenOct(int index, String text)
    {
        int i = Integer.parseInt(text, 8);
        return new Token(index, yyline, yycolumn, yytext(), new Integer(i));
    }

    private Token newTokenOct(int index)
    {
        return newTokenOct(index, yytext());
    }

    private Token newTokenHex(int index, String text)
    {
        int i;
        try {
            i = Integer.parseInt(text, 16);
        }
        catch (NumberFormatException nfe) {
            // XXX could also be negative, php converts to double
            System.out.println("Hex number id too big: " + text);
            i = Integer.MAX_VALUE;
        }
        return new Token(index, yyline, yycolumn, yytext(), new Integer(i));
    }

    private Token newTokenHex(int index)
    {
        return newTokenHex(index, yytext());
    }

    private Token newTokenDouble(int index, String text)
    {
        return new Token(index, yyline, yycolumn, yytext(), new Double(text));
    }

    private Token newTokenDouble(int index)
    {
        return newTokenDouble(index, yytext());
    }

    private Token newTokenDoubleQuoteString(int index, String text)
    {
        return new Token(index, yyline, yycolumn, yytext(), StringTools.decodeDoubleQuoteString(text));
    }

    private Token newTokenDoubleQuoteString(int index)
    {
        return newTokenDoubleQuoteString(index, yytext());
    }

    private Token newTokenSingleQuoteString(int index, String text)
    {
        return new Token(index, yyline, yycolumn, yytext(), StringTools.decodeSingleQuoteString(text));
    }

    private Token newTokenSingleQuoteString(int index)
    {
        return newTokenSingleQuoteString(index, yytext());
    }

    public antlr.Token nextToken() throws antlr.TokenStreamException
    {
        try {
            antlr.Token token = null;
            int         type  = 0;
            do {
                token = yylex();
                type  = token.getType();
            }
            while (type == Token.T_COMMENT || type == Token.T_DOC_COMMENT || type == Token.T_OPEN_TAG || type == Token.T_WHITESPACE);
            /*
            if (type == Token.T_COMMENT || type == Token.T_DOC_COMMENT || type == Token.T_OPEN_TAG || type == Token.T_WHITESPACE) {
                token = nextToken();
            }
            */
            if (type == Token.T_CLOSE_TAG) {
                token.setType(Token.T_SEMI_COLON);
            }
            else if (type == Token.T_OPEN_TAG_WITH_ECHO) {
                token.setType(Token.T_ECHO);
            }
            return token;
        }
        catch (java.io.IOException e){
            antlr.TokenStreamException tse = new antlr.TokenStreamException();
            tse.initCause(e);
            throw tse;
        }
    }
%}

%x ST_IN_SCRIPTING
%x ST_DOUBLE_QUOTES
%x ST_SINGLE_QUOTE
%x ST_BACKQUOTE
%x ST_HEREDOC
%x ST_LOOKING_FOR_PROPERTY
%x ST_LOOKING_FOR_VARNAME
%x ST_COMMENT
%x ST_DOC_COMMENT
%x ST_ONE_LINE_COMMENT

%public
%class PHPScanner
%implements antlr.TokenStream
//%function nextToken
//%standalone
%type antlr.Token
//%debug
%char
%line
%column
%unicode
%ignorecase

%eofval{
    return newToken(Token.EOF_TYPE);
%eofval}

LNUM                   = [0-9]+
DNUM                   = ([0-9]*[\.][0-9]+)|([0-9]+[\.][0-9]*)
EXPONENT_DNUM          = (({LNUM}|{DNUM})[eE][+-]?{LNUM})
HNUM                   = "0x"[0-9a-fA-F]+
LABEL                  = [a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*
WHITESPACE             = [ \n\r\t]+
TABS_AND_SPACES        = [ \t]*
TOKENS                 = [;:,.\[\]()|\^&+-/*=%!~\$<>?@]
// '$' seems to be unneccessary because its also matched by "$"[^a-z...]
ENCAPSED_TOKENS        = [\[\]{}\$]
ESCAPED_AND_WHITESPACE = [\n\t\r #'.:;,()|\^&+-/*=%!~<>?@]+
ANY_CHAR               = (.|[\n])
NEWLINE                = ("\r"|"\n"|"\r\n")

%%

<ST_IN_SCRIPTING>"exit" {
    return newToken(Token.T_EXIT);
}

<ST_IN_SCRIPTING>"die" {
    return newToken(Token.T_EXIT);
}

<ST_IN_SCRIPTING>"function" {
    return newToken(Token.T_FUNCTION);
}

<ST_IN_SCRIPTING>"const" {
    return newToken(Token.T_CONST);
}

<ST_IN_SCRIPTING>"return" {
    return newToken(Token.T_RETURN);
}

<ST_IN_SCRIPTING>"try" {
    return newToken(Token.T_TRY);
}

<ST_IN_SCRIPTING>"catch" {
    return newToken(Token.T_CATCH);
}

<ST_IN_SCRIPTING>"throw" {
    return newToken(Token.T_THROW);
}

<ST_IN_SCRIPTING>"if" {
    return newToken(Token.T_IF);
}

<ST_IN_SCRIPTING>"elseif" {
    return newToken(Token.T_ELSEIF);
}

<ST_IN_SCRIPTING>"endif" {
    return newToken(Token.T_ENDIF);
}

<ST_IN_SCRIPTING>"else" {
    return newToken(Token.T_ELSE);
}

<ST_IN_SCRIPTING>"while" {
    return newToken(Token.T_WHILE);
}

<ST_IN_SCRIPTING>"endwhile" {
    return newToken(Token.T_ENDWHILE);
}

<ST_IN_SCRIPTING>"do" {
    return newToken(Token.T_DO);
}

<ST_IN_SCRIPTING>"for" {
    return newToken(Token.T_FOR);
}

<ST_IN_SCRIPTING>"endfor" {
    return newToken(Token.T_ENDFOR);
}

<ST_IN_SCRIPTING>"foreach" {
    return newToken(Token.T_FOREACH);
}

<ST_IN_SCRIPTING>"endforeach" {
    return newToken(Token.T_ENDFOREACH);
}

<ST_IN_SCRIPTING>"declare" {
    return newToken(Token.T_DECLARE);
}

<ST_IN_SCRIPTING>"enddeclare" {
    return newToken(Token.T_ENDDECLARE);
}

<ST_IN_SCRIPTING>"instanceof" {
    return newToken(Token.T_INSTANCEOF);
}

<ST_IN_SCRIPTING>"as" {
    return newToken(Token.T_AS);
}

<ST_IN_SCRIPTING>"switch" {
    return newToken(Token.T_SWITCH);
}

<ST_IN_SCRIPTING>"endswitch" {
    return newToken(Token.T_ENDSWITCH);
}

<ST_IN_SCRIPTING>"case" {
    return newToken(Token.T_CASE);
}

<ST_IN_SCRIPTING>"default" {
    return newToken(Token.T_DEFAULT);
}

<ST_IN_SCRIPTING>"break" {
    return newToken(Token.T_BREAK);
}

<ST_IN_SCRIPTING>"continue" {
    return newToken(Token.T_CONTINUE);
}

<ST_IN_SCRIPTING>"echo" {
    return newToken(Token.T_ECHO);
}

<ST_IN_SCRIPTING>"print" {
    return newToken(Token.T_PRINT);
}

<ST_IN_SCRIPTING>"class" {
    return newToken(Token.T_CLASS);
}

<ST_IN_SCRIPTING>"interface" {
    return newToken(Token.T_INTERFACE);
}

<ST_IN_SCRIPTING>"extends" {
    return newToken(Token.T_EXTENDS);
}

<ST_IN_SCRIPTING>"implements" {
    return newToken(Token.T_IMPLEMENTS);
}

<ST_IN_SCRIPTING,ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_HEREDOC>"->" {
    yypushstate(ST_LOOKING_FOR_PROPERTY);
    return newToken(Token.T_OBJECT_OPERATOR);
}

<ST_LOOKING_FOR_PROPERTY>{LABEL} {
    yypopstate();
    //zend_copy_value(zendlval, yytext, yyleng);
    //zendlval->type = IS_STRING;
    return newTokenText(Token.T_STRING);
}

<ST_LOOKING_FOR_PROPERTY>{ANY_CHAR} {
    yypushback(1);
    //yyless(0);
    yypopstate();
}

<ST_IN_SCRIPTING>"::" {
    return newToken(Token.T_PAAMAYIM_NEKUDOTAYIM);
}

<ST_IN_SCRIPTING>"new" {
    return newToken(Token.T_NEW);
}

<ST_IN_SCRIPTING>"clone" {
    return newToken(Token.T_CLONE);
}

<ST_IN_SCRIPTING>"var" {
    return newToken(Token.T_VAR);
}

<ST_IN_SCRIPTING>"("{TABS_AND_SPACES}("int"|"integer"){TABS_AND_SPACES}")" {
    return newToken(Token.T_INT_CAST);
}

<ST_IN_SCRIPTING>"("{TABS_AND_SPACES}("real"|"double"|"float"){TABS_AND_SPACES}")" {
    return newToken(Token.T_DOUBLE_CAST);
}

<ST_IN_SCRIPTING>"("{TABS_AND_SPACES}"string"{TABS_AND_SPACES}")" {
    return newToken(Token.T_STRING_CAST);
}

<ST_IN_SCRIPTING>"("{TABS_AND_SPACES}"array"{TABS_AND_SPACES}")" {
    return newToken(Token.T_ARRAY_CAST);
}

<ST_IN_SCRIPTING>"("{TABS_AND_SPACES}"object"{TABS_AND_SPACES}")" {
    return newToken(Token.T_OBJECT_CAST);
}

<ST_IN_SCRIPTING>"("{TABS_AND_SPACES}("bool"|"boolean"){TABS_AND_SPACES}")" {
    return newToken(Token.T_BOOL_CAST);
}

<ST_IN_SCRIPTING>"("{TABS_AND_SPACES}("unset"){TABS_AND_SPACES}")" {
    return newToken(Token.T_UNSET_CAST);
}

<ST_IN_SCRIPTING>"eval" {
    return newToken(Token.T_EVAL);
}

<ST_IN_SCRIPTING>"include" {
    return newToken(Token.T_INCLUDE);
}

<ST_IN_SCRIPTING>"include_once" {
    return newToken(Token.T_INCLUDE_ONCE);
}

<ST_IN_SCRIPTING>"require" {
    return newToken(Token.T_REQUIRE);
}

<ST_IN_SCRIPTING>"require_once" {
    return newToken(Token.T_REQUIRE_ONCE);
}

<ST_IN_SCRIPTING>"use" {
    return newToken(Token.T_USE);
}

<ST_IN_SCRIPTING>"global" {
    return newToken(Token.T_GLOBAL);
}

<ST_IN_SCRIPTING>"isset" {
    return newToken(Token.T_ISSET);
}

<ST_IN_SCRIPTING>"empty" {
    return newToken(Token.T_EMPTY);
}

<ST_IN_SCRIPTING>"static" {
    return newToken(Token.T_STATIC);
}

<ST_IN_SCRIPTING>"abstract" {
    return newToken(Token.T_ABSTRACT);
}

<ST_IN_SCRIPTING>"final" {
    return newToken(Token.T_FINAL);
}

<ST_IN_SCRIPTING>"private" {
    return newToken(Token.T_PRIVATE);
}

<ST_IN_SCRIPTING>"protected" {
    return newToken(Token.T_PROTECTED);
}

<ST_IN_SCRIPTING>"public" {
    return newToken(Token.T_PUBLIC);
}

<ST_IN_SCRIPTING>"unset" {
    return newToken(Token.T_UNSET);
}

<ST_IN_SCRIPTING>"=>" {
    return newToken(Token.T_DOUBLE_ARROW);
}

<ST_IN_SCRIPTING>"list" {
    return newToken(Token.T_LIST);
}

<ST_IN_SCRIPTING>"array" {
    return newToken(Token.T_ARRAY);
}

<ST_IN_SCRIPTING>"++" {
    return newToken(Token.T_INC);
}

<ST_IN_SCRIPTING>"--" {
    return newToken(Token.T_DEC);
}

<ST_IN_SCRIPTING>"===" {
    return newToken(Token.T_IS_IDENTICAL);
}

<ST_IN_SCRIPTING>"!==" {
    return newToken(Token.T_IS_NOT_IDENTICAL);
}

<ST_IN_SCRIPTING>"==" {
    return newToken(Token.T_IS_EQUAL);
}

<ST_IN_SCRIPTING>"!="|"<>" {
    return newToken(Token.T_IS_NOT_EQUAL);
}

<ST_IN_SCRIPTING>"<=" {
    return newToken(Token.T_IS_SMALLER_OR_EQUAL);
}

<ST_IN_SCRIPTING>">=" {
    return newToken(Token.T_IS_GREATER_OR_EQUAL);
}

<ST_IN_SCRIPTING>"+=" {
    return newToken(Token.T_PLUS_EQUAL);
}

<ST_IN_SCRIPTING>"-=" {
    return newToken(Token.T_MINUS_EQUAL);
}

<ST_IN_SCRIPTING>"*=" {
    return newToken(Token.T_MUL_EQUAL);
}

<ST_IN_SCRIPTING>"/=" {
    return newToken(Token.T_DIV_EQUAL);
}

<ST_IN_SCRIPTING>".=" {
    return newToken(Token.T_CONCAT_EQUAL);
}

<ST_IN_SCRIPTING>"%=" {
    return newToken(Token.T_MOD_EQUAL);
}

<ST_IN_SCRIPTING>"<<=" {
    return newToken(Token.T_SL_EQUAL);
}

<ST_IN_SCRIPTING>">>=" {
    return newToken(Token.T_SR_EQUAL);
}

<ST_IN_SCRIPTING>"&=" {
    return newToken(Token.T_AND_EQUAL);
}

<ST_IN_SCRIPTING>"|=" {
    return newToken(Token.T_OR_EQUAL);
}

<ST_IN_SCRIPTING>"^=" {
    return newToken(Token.T_XOR_EQUAL);
}

<ST_IN_SCRIPTING>"||" {
    return newToken(Token.T_BOOLEAN_OR);
}

<ST_IN_SCRIPTING>"&&" {
    return newToken(Token.T_BOOLEAN_AND);
}

<ST_IN_SCRIPTING>"OR" {
    return newToken(Token.T_LOGICAL_OR);
}

<ST_IN_SCRIPTING>"AND" {
    return newToken(Token.T_LOGICAL_AND);
}

<ST_IN_SCRIPTING>"XOR" {
    return newToken(Token.T_LOGICAL_XOR);
}

<ST_IN_SCRIPTING>"<<" {
    return newToken(Token.T_SL);
}

<ST_IN_SCRIPTING>">>" {
    return newToken(Token.T_SR);
}

<ST_IN_SCRIPTING>{TOKENS} {
    //return yytext[0];
    return newToken(yycharat(0));
}

<ST_IN_SCRIPTING>"{" {
    yypushstate(ST_IN_SCRIPTING);
    return newToken('{');
}

<ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_HEREDOC>"${" {
    yypushstate(ST_LOOKING_FOR_VARNAME);
    return newToken(Token.T_DOLLAR_OPEN_CURLY_BRACES);
}

<ST_IN_SCRIPTING>"}" {
    yypopstate();
    return newToken('}');
}

<ST_LOOKING_FOR_VARNAME>{LABEL} {
    //zend_copy_value(zendlval, yytext, yyleng);
    //zendlval->type = IS_STRING;
    yypopstate();
    yypushstate(ST_IN_SCRIPTING);
    return newTokenText(Token.T_STRING_VARNAME);
}

<ST_LOOKING_FOR_VARNAME>{ANY_CHAR} {
    //yyless(0);
    yypushback(1);
    yypopstate();
    yypushstate(ST_IN_SCRIPTING);
}

<ST_IN_SCRIPTING>{LNUM} {
    return newTokenInt(Token.T_LNUMBER);
}

<ST_IN_SCRIPTING>{HNUM} {
    return newTokenHex(Token.T_LNUMBER);
}

<ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_HEREDOC>{LNUM}|{HNUM} { /* treat numbers (almost) as strings inside encapsulated strings */
    //zendlval->value.str.val = (char *)estrndup(yytext, yyleng);
    //zendlval->value.str.len = yyleng;
    //zendlval->type = IS_STRING;
    return newTokenText(Token.T_NUM_STRING);
}

<ST_IN_SCRIPTING>{DNUM}|{EXPONENT_DNUM} {
    //zendlval->value.dval = zend_strtod(yytext, NULL);
    //zendlval->type = IS_DOUBLE;
    return newTokenDouble(Token.T_DNUMBER);
}

<ST_IN_SCRIPTING>"__CLASS__" {
    return newToken(Token.T_CLASS_C);
}

<ST_IN_SCRIPTING>"__FUNCTION__" {
    return newToken(Token.T_FUNC_C);
}

<ST_IN_SCRIPTING>"__METHOD__" {
    return newToken(Token.T_METHOD_C);
}

<ST_IN_SCRIPTING>"__LINE__" {
    return newToken(Token.T_LINE);
}

<ST_IN_SCRIPTING>"__FILE__" {
    return newToken(Token.T_FILE);
}

<YYINITIAL>(([^<]|"<"[^?%s<]){1,400})|"<s"|"<" {
    return newTokenText(Token.T_INLINE_HTML);
}

<YYINITIAL>"<?"|"<script"{WHITESPACE}+"language"{WHITESPACE}*"="{WHITESPACE}*("php"|"\"php\""|"\'php\'"){WHITESPACE}*">" {
    if (ENABLE_SHORT_TAGS || yylength() > 2) { /* yyleng>2 means it's not <? but <script> */
        yybegin(ST_IN_SCRIPTING);
        return newToken(Token.T_OPEN_TAG);
    }
    else {
        return newTokenText(Token.T_INLINE_HTML);
    }
}

<YYINITIAL>"<%="|"<?=" {
    if ((yycharat(1) == '%' && ENABLE_ASP_TAGS) || (yycharat(1) == '?' && ENABLE_SHORT_TAGS)) {
        yybegin(ST_IN_SCRIPTING);
        return newToken(Token.T_OPEN_TAG_WITH_ECHO);
    }
    else {
        return newTokenText(Token.T_INLINE_HTML);
    }
}

<YYINITIAL>"<%" {
    if (ENABLE_ASP_TAGS) {
        yybegin(ST_IN_SCRIPTING);
        return newToken(Token.T_OPEN_TAG);
    }
    else {
        return newTokenText(Token.T_INLINE_HTML);
    }
}

<YYINITIAL>"<?php"([ \t]|{NEWLINE}) {
    yybegin(ST_IN_SCRIPTING);
    return newToken(Token.T_OPEN_TAG);
}

<ST_IN_SCRIPTING,ST_DOUBLE_QUOTES,ST_HEREDOC,ST_BACKQUOTE>"$"{LABEL} {
    //zend_copy_value(zendlval, (yytext+1), (yyleng-1));
    //zendlval->type = IS_STRING;
    return newTokenText(Token.T_VARIABLE, yytext().substring(1));
}

<ST_IN_SCRIPTING>{LABEL} {
    //zend_copy_value(zendlval, yytext, yyleng);
    //zendlval->type = IS_STRING;
    return newTokenText(Token.T_STRING);
}

<ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_HEREDOC>{LABEL} {
    //zend_copy_value(zendlval, yytext, yyleng);
    //zendlval->type = IS_STRING;
    return newTokenText(Token.T_STRING);
}

<ST_IN_SCRIPTING>{WHITESPACE} {
    return newToken(Token.T_WHITESPACE);
}

<ST_IN_SCRIPTING>"#"|"//" {
    comment.append(yytext());
    yybegin(ST_ONE_LINE_COMMENT);
}

<ST_ONE_LINE_COMMENT>"?"|"%"|">" {
    comment.append(yytext());
    //yymore();
}

<ST_ONE_LINE_COMMENT>[^\n\r?%>]+ {
    comment.append(yytext());
    //yymore();
}

<ST_ONE_LINE_COMMENT>{NEWLINE} {
    comment.append(yytext());
    String text = comment.toString();
    comment.setLength(0);
    yybegin(ST_IN_SCRIPTING);
    return newTokenText(Token.T_COMMENT, text);
}

<ST_ONE_LINE_COMMENT>"?>"|"%>" {
    if (ENABLE_ASP_TAGS || yycharat(0) != '%') {
        String text = comment.toString();
        comment.setLength(0);
        yybegin(ST_IN_SCRIPTING);
        yypushback(2);
        return newTokenText(Token.T_COMMENT, text);
    }
    else {
        comment.append(yytext());
    }
}

<ST_IN_SCRIPTING>"/**"{WHITESPACE} {
    comment.append(yytext());
    yybegin(ST_DOC_COMMENT);
}

<ST_IN_SCRIPTING>"/*" {
    comment.append(yytext());
    yybegin(ST_COMMENT);
    //yymore();
}

<ST_COMMENT,ST_DOC_COMMENT>[^*]+ {
    comment.append(yytext());
    //yymore();
}

<ST_DOC_COMMENT>"*/" {
    comment.append(yytext());
    String text = comment.toString();
    comment.setLength(0);
    yybegin(ST_IN_SCRIPTING);
    return newTokenText(Token.T_DOC_COMMENT, text);
}

<ST_COMMENT>"*/" {
    comment.append(yytext());
    String text = comment.toString();
    comment.setLength(0);
    yybegin(ST_IN_SCRIPTING);
    return newTokenText(Token.T_COMMENT, text);
}

<ST_COMMENT,ST_DOC_COMMENT>"*" {
    comment.append(yytext());
    //yymore();
}

<ST_IN_SCRIPTING>("?>"|"</script"{WHITESPACE}*">"){NEWLINE}? {
    yybegin(YYINITIAL);
    return newToken(Token.T_CLOSE_TAG);  /* implicit ';' at php-end tag */
}

<ST_IN_SCRIPTING>"%>"{NEWLINE}? {
    if (ENABLE_ASP_TAGS) {
        return newToken(Token.T_CLOSE_TAG);
    }
    else {
        yypushback(yylength()-1);
        return newToken(yycharat(0));
    }
}

<ST_IN_SCRIPTING>([\"]([^\$\"\\]|("\\".))*[\"]) {
    return newTokenDoubleQuoteString(Token.T_CONSTANT_ENCAPSED_STRING, yytext().substring(1, yylength()-1));
}

<ST_IN_SCRIPTING>([']([^'\\]|("\\".))*[']) {
    return newTokenSingleQuoteString(Token.T_CONSTANT_ENCAPSED_STRING, yytext().substring(1, yylength()-1));
}

<ST_IN_SCRIPTING>[\"] {
    yybegin(ST_DOUBLE_QUOTES);
    return newToken('\"');
}

<ST_IN_SCRIPTING>"<<<"{TABS_AND_SPACES}{LABEL}{NEWLINE} {
    yybegin(ST_HEREDOC);
    heredocLabel = yytext().substring(3).trim();
    //System.out.println(heredocLabel);
    return newTokenText(Token.T_START_HEREDOC, heredocLabel);
}

<ST_IN_SCRIPTING>[`] {
    yybegin(ST_BACKQUOTE);
    return newToken('`');
}

<ST_IN_SCRIPTING>['] {
    yybegin(ST_SINGLE_QUOTE);
    return newToken('\'');
}

<ST_HEREDOC>^{LABEL}(";")?{NEWLINE} {
    String label = yytext();
    int    endindex = label.length()-1;
    char   ch       = label.charAt(endindex);
    boolean unput   = false;
    while (ch == '\r' || ch == '\n') {
        ch = label.charAt(--endindex);
    }
    if (label.charAt(endindex) == ';') {
        --endindex;
        unput = true;
    }
    label = label.substring(0, endindex+1);
    if (label.equals(heredocLabel)) {
        if (unput) {
            zzBuffer[--zzMarkedPos] = ';';
        }
        yybegin(ST_IN_SCRIPTING);
        return newTokenText(Token.T_END_HEREDOC, label);
    }
    else {
        return newTokenText(Token.T_STRING);
    }
}

<ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_HEREDOC>{ESCAPED_AND_WHITESPACE} {
    return newTokenText(Token.T_ENCAPSED_AND_WHITESPACE);
}

<ST_SINGLE_QUOTE>([^'\\]|\\[^'\\])+ {
    return newTokenText(Token.T_ENCAPSED_AND_WHITESPACE);
}

<ST_DOUBLE_QUOTES>[`]+ {
    return newTokenText(Token.T_ENCAPSED_AND_WHITESPACE);
}

<ST_BACKQUOTE>[\"]+ {
    return newTokenText(Token.T_ENCAPSED_AND_WHITESPACE);
}

<ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_HEREDOC>"$"[^a-zA-Z_\x7f-\xff{] {
    if (yylength() == 2) {
        yypushback(1);
    }
    return newTokenText(Token.T_CHARACTER);
}

<ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_HEREDOC>{ENCAPSED_TOKENS} {
    return newTokenText(yycharat(0));
}

<ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_HEREDOC>"{$" {
    yypushback(1);
    return newToken(Token.T_CURLY_OPEN);
}

<ST_SINGLE_QUOTE>"\\'" {
    return newTokenText(Token.T_CHARACTER);
}

<ST_SINGLE_QUOTE>"\\\\" {
    return newTokenText(Token.T_CHARACTER);
}

<ST_DOUBLE_QUOTES>"\\\"" {
    return newTokenText(Token.T_CHARACTER);
}

<ST_BACKQUOTE>"\\`" {
    return newTokenText(Token.T_CHARACTER);
}

<ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_HEREDOC>"\\"[0-7]{1,3} {
    return newTokenOct(Token.T_CHARACTER, yytext().substring(1));
}

<ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_HEREDOC>"\\x"[0-9A-Fa-f]{1,2} {
    return newTokenHex(Token.T_CHARACTER, yytext().substring(2));
}

<ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_HEREDOC>"\\"{ANY_CHAR} {
    char ch = yycharat(1);
    switch (ch) {
        case 'r': return newTokenText(Token.T_CHARACTER, "\r");
        case 'n': return newTokenText(Token.T_CHARACTER, "\n");
        case 't': return newTokenText(Token.T_CHARACTER, "\t");
        case '$': return newTokenText(Token.T_CHARACTER, "$");
        case '{': return newTokenText(Token.T_CHARACTER, "{");
        case '\\': return newTokenText(Token.T_CHARACTER, "\\");
        default : return newTokenText(Token.T_BAD_CHARACTER, "\\" + ch);
    }
    //return newToken(Token.T_CHARACTER);
}

<ST_HEREDOC>[\"'`]+ {
    return newTokenText(Token.T_ENCAPSED_AND_WHITESPACE, yytext());
}

<ST_DOUBLE_QUOTES>[\"] {
    yybegin(ST_IN_SCRIPTING);
    return newToken('\"');
}

<ST_BACKQUOTE>[`] {
    yybegin(ST_IN_SCRIPTING);
    return newToken('`');
}

<ST_SINGLE_QUOTE>['] {
    yybegin(ST_IN_SCRIPTING);
    return newToken('\'');
}

// XXX is this needed?
<ST_DOUBLE_QUOTES,ST_BACKQUOTE,YYINITIAL,ST_IN_SCRIPTING,ST_LOOKING_FOR_PROPERTY><<EOF>> {
    return newToken(Token.EOF_TYPE);
    //return null;
}

<ST_COMMENT,ST_DOC_COMMENT><<EOF>> {
    System.out.println("Unterminated comment starting line " + yyline);
    return newToken(Token.EOF_TYPE);
}

<ST_IN_SCRIPTING,YYINITIAL,ST_DOUBLE_QUOTES,ST_BACKQUOTE,ST_SINGLE_QUOTE,ST_HEREDOC>{ANY_CHAR} {
    System.out.println("Unexpected character in input: " + yytext());
}

