public class StringTools
{
    private StringTools() { }

    public static String decodeDoubleQuoteString(String in)
    {
        int    i   = 0;
        int    j   = 0;
        int    tmp = 0;
        int    len = in.length();
        char[] out = new char[len];
        char   ch;
        while (i < len) {
            ch = in.charAt(i++);
            if (ch == '\\') {
                if (i >= len) {
                    continue;
                }
                ch = in.charAt(i++);
                switch (ch) {
                    case 'n':
                        out[j++] = '\n';
                        break;
                    case 'r':
                        out[j++] = '\r';
                        break;
                    case 't':
                        out[j++] = '\t';
                        break;
                    case '\\':
                    case '$':
                    case '"':
                        out[j++] = ch;
                        break;
                    case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7':
                        tmp = ch - '0';
                        if (i < len) {
                            ch = in.charAt(i++);
                            if (ch >= '0' && ch <= '7') {
                                tmp = tmp*8 + ch - '0';
                                if (i < len) {
                                    ch = in.charAt(i++);
                                    if (ch >= '0' && ch <= '7') {
                                        tmp = tmp*8 + ch - '0';
                                        ch  = in.charAt(i++);
                                    }
                                }
                            }
                        }
                        out[j++] = (char)tmp;
                        break;
                    case 'x':
                        if (i < len) {
                            ch = in.charAt(i++);
                            if (ch >= '0' && ch <= '9' || ch >= 'a' && ch <= 'f' || ch >= 'A' && ch <= 'F') {
                                tmp = tmp*16 + Character.digit(ch, 16);
                                if (i < len) {
                                    ch = in.charAt(i++);
                                    if (ch >= '0' && ch <= '9' || ch >= 'a' && ch <= 'f' || ch >= 'A' && ch <= 'F') {
                                        tmp = tmp*16 + Character.digit(ch, 16);
                                    }
                                }
                                out[j++] = (char)tmp;
                            }
                            else {
                                out[j++] = '\\';
                                out[j++] = ch;
                            }
                        }
                        break;
                    default:
                        out[j++] = '\\';
                        out[j++] = ch;
                        break;
                }
            }
            else {
                out[j++] = ch;
            }
        }
        return new String(out, 0, j);
    }

    public static String decodeSingleQuoteString(String in)
    {
        int    i   = 0;
        int    j   = 0;
        int    tmp = 0;
        int    len = in.length();
        char[] out = new char[len];
        char   ch;
        while (i < len) {
            ch = in.charAt(i++);
            if (ch == '\\') {
                if (i >= len) {
                    continue;
                }
                ch = in.charAt(i++);
                switch (ch) {
                    case '\\':
                    case '\'':
                        out[j++] = ch;
                        break;
                    default:
                        out[j++] = '\\';
                        out[j++] = ch;
                        break;
                }
            }
            else {
                out[j++] = ch;
            }
        }
        return new String(out, 0, j);
    }

    public static String escapeString(String in)
    {
        int          i   = 0;
        int          len = in.length();
        StringBuffer sb  = new StringBuffer(len);
        char         ch;
        while (i < len) {
            ch = in.charAt(i++);
            if (ch < 32) {
                switch (ch) {
                    case '\r':
                        sb.append("\\r");
                        break;
                    case '\n':
                        sb.append("\\n");
                        break;
                    case '\t':
                        sb.append("\\t");
                        break;
                    default:
                        sb.append("\\x");
                        sb.append(Integer.toHexString(ch));
                        break;
                }
            }
            else {
                sb.append(ch);
            }
        }
        return sb.toString();
    }

    public static void main(String[] args)
    {
        System.out.println(escapeString(decodeDoubleQuoteString("test\\r\\n\\$asdf\\x09\\\\")));
        System.out.println(escapeString(decodeDoubleQuoteString("a\\\\b\\\\c\\\\")));
        System.out.println(escapeString(decodeDoubleQuoteString("x")));
        System.out.println(escapeString(decodeDoubleQuoteString("abc")));
        System.out.println(escapeString(decodeDoubleQuoteString("a")));
        System.out.println(escapeString(decodeDoubleQuoteString("\\t")));
        System.out.println(escapeString(decodeDoubleQuoteString("\\ta")));
        System.out.println(escapeString(decodeDoubleQuoteString("\\\\\\")));
        System.out.println(escapeString(decodeDoubleQuoteString("\\\"")));

        System.out.println(escapeString(decodeDoubleQuoteString("\"a\"")));
        System.out.println(escapeString(decodeDoubleQuoteString("\"a\"".substring(1,2))));

        System.out.println(escapeString(decodeDoubleQuoteString("\\x1b")));
        System.out.println(escapeString(decodeDoubleQuoteString("\\x7e")));

        System.out.println(escapeString(decodeSingleQuoteString("\\abc")));
        System.out.println(escapeString(decodeSingleQuoteString("\\'")));
    }
}
